/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc-test:fr.bigeon.gclc.test.TestConsoleManager.java
 * Created on: Nov 18, 2016
 */
package fr.bigeon.gclc.manager;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.nio.charset.Charset;

/** This console manager allows to enter commands and retrieve the output as an
 * input.
 * <p>
 * This console manager is used to internally pilot an application. This can be
 * used to test application behavior.
 *
 * @author Emmanuel Bigeon */
public final class PipedConsoleManager
        implements ConsoleManager {

    /** The encoding between streams */
    private static final String UTF_8 = "UTF-8"; //$NON-NLS-1$
    /** THe inner manager */
    private final SystemConsoleManager innerManager;
    /** The stream to pipe commands into */
    private final PipedOutputStream commandInput;
    /** The reader to get application return from */
    private final BufferedReader commandBuffOutput;
    /** The stream to get application return from */
    private final PipedInputStream commandOutput;
    /** The print writer for application to write return to */
    private final PrintStream outPrint;
    /** The stream for the application to read commands from */
    private final PipedInputStream in;
    /** The writing thread */
    private final WritingRunnable writing;
    /** The reading thread */
    private final ReadingRunnable reading;

    /** @throws IOException if the piping failed for streams */
    public PipedConsoleManager() throws IOException {
        commandInput = new PipedOutputStream();
        in = new PipedInputStream(commandInput);
        commandOutput = new PipedInputStream();
        PipedOutputStream out = new PipedOutputStream(commandOutput);
        commandBuffOutput = new BufferedReader(
                new InputStreamReader(commandOutput, Charset.forName(UTF_8)));
        outPrint = new PrintStream(out, true, UTF_8);
        innerManager = new SystemConsoleManager(outPrint, in,
                Charset.forName(UTF_8));
        writing = new WritingRunnable(commandInput, Charset.forName(UTF_8));
        reading = new ReadingRunnable(commandBuffOutput);
        Thread th = new Thread(writing, "write"); //$NON-NLS-1$
        th.start();
        th = new Thread(reading, "read"); //$NON-NLS-1$
        th.setDaemon(true);
        th.start();
    }

    @Override
    public String getPrompt() {
        return innerManager.getPrompt();
    }

    @Override
    public void print(String object) throws IOException {
        innerManager.print(object);
    }

    @Override
    public void println() throws IOException {
        innerManager.println();
    }

    @Override
    public void println(String object) throws IOException {
        innerManager.println(object);
    }

    @Override
    public String prompt() throws IOException {
        return innerManager
                .prompt(innerManager.getPrompt() + System.lineSeparator());
    }

    @Override
    public String prompt(String message) throws IOException {
        return innerManager.prompt(message + System.lineSeparator());
    }

    @Override
    public void setPrompt(String prompt) {
        innerManager.setPrompt(prompt);
    }

    @Override
    public void close() throws IOException {
        innerManager.close();
        reading.setRunning(false);
        writing.setRunning(false);
        outPrint.close();
        commandBuffOutput.close();
        commandOutput.close();
        in.close();
        commandInput.close();
    }

    @Override
    public boolean isClosed() {
        return innerManager.isClosed();
    }

    /** @param content the content to type to the application
     * @throws IOException if the typing failed */
    public void type(String content) throws IOException {
        writing.addMessage(content);
    }

    /** @return the content of the next line written by the application
     * @throws IOException if the reading failed */
    public String readNextLine() throws IOException {
        return reading.getMessage();
    }

    /** @return the content of the next line written by the application
     * @throws IOException if the reading failed */
    public boolean available() throws IOException {
        return reading.hasMessage();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleManager#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        innerManager.interruptPrompt();
    }

    /** @param message the message
     * @return the thread to join to wait for message delivery
     * @see fr.bigeon.gclc.manager.ReadingRunnable#getWaitForDelivery(java.lang.String) */
    public Thread getWaitForDelivery(String message) {
        return reading.getWaitForDelivery(message);
    }

}
